/*
  Motion Puttern Builder
  Sample for Arduino
             ESP32-WROOM / ESP32-WROVER

  NIPPON PULSE MOTOR Co., Ltd.
*/

#include <Wire.h>
#include "LCD_control.h"

//----------------------------------------------
//I2C transfer ---------------------------------
// Terminal 
#define lcd_SDA  21
#define lcd_SCL  22
//I2C speed
static const uint32_t lcd_I2C_CLK = 100000; // I2C = 100k(10us)

byte contrast = 26;   // コントラスト(0～63)
bool TopOfSending = true; // I2C transfer control


////////////////////////////////////////////////////////////////
// write command to LCD
//
uint8_t LCD_CONTROL::write_cmd(uint8_t command, uint8_t data, uint32_t wait_ms)
{
  uint8_t ret = 0;
  //
  if(TopOfSending == true)
  {
    Wire.beginTransmission(LCD_ADRS);
  }
  Wire.write(command);
  Wire.write(data);
  //
  if((command & 0x0080) == 0)
  {
    // This send is last
    ret = Wire.endTransmission();
    TopOfSending = true;
  }
  else
  {
    // Still sending
    TopOfSending = false;
  }
  //
  if(wait_ms > 0)
  {
    delay(wait_ms);
  }
  //
  return (ret);
}


////////////////////////////////////////////////////////////////
// LCD initial
//
void LCD_CONTROL::begin(void)
{ 
  Wire.begin(lcd_SDA, lcd_SCL);
  Wire.setClock(lcd_I2C_CLK);
  //
  write_cmd(0x00, 0x01, 2);//clear
  //
  write_cmd(0x00, 0x38, 0);
  write_cmd(0x00, 0x39, 0);//2-line x1
  write_cmd(0x00, 0x14, 0);
  write_cmd(0x00, (0x70 | (contrast & 0xF)), 0);
  write_cmd(0x00, (0x5C | ((contrast>>4) &0x3)), 0);
  write_cmd(0x00, 0x6C, 201);
  write_cmd(0x00, 0x38, 0);//2-line x1
  write_cmd(0x00, 0x01, 2);//clear
  write_cmd(0x00, 0x0C, 0);
}


////////////////////////////////////////////////////////////////
// 
//
void LCD_CONTROL::printString(String pdata)
{
  int n = pdata.length() - 1;
  for (int i = 0; i < n; i = i + 1)
  {
    write_cmd(0xC0, pdata.charAt(i), 1);
  }
  write_cmd(0x40, pdata.charAt(n), 1);
}


////////////////////////////////////////////////////////////////
// LCD clear
//
void LCD_CONTROL::clear(void)
{
  write_cmd(0x00, 0x01, 0);
} 


////////////////////////////////////////////////////////////////
// LCD display off
//
void LCD_CONTROL::DisplayOff()
{
  write_cmd(0x00, (0x08 < 0), 0);
}


////////////////////////////////////////////////////////////////
// LCD display on
//
void LCD_CONTROL::DisplayOn()
{
  write_cmd(0x00, 0x0C, 0);
}


////////////////////////////////////////////////////////////////
// Cursor shift
//
void LCD_CONTROL::shift_cursor(uint8_t cnt, uint8_t shift_cmd)
{
  uint8_t a;
  for(a = 0; a < cnt; a ++)
  {
    write_cmd(0x00, shift_cmd, 0);
  }
}


////////////////////////////////////////////////////////////////
// 
//
void LCD_CONTROL::Display(String st1, String st2, bool ini)
{
  if(ini == true)
  {
    clear(); 
    write_cmd(0x00, 0x02, 2);//30); // Return Home
  }
  printString(st1);
  if(st2 != "")
  {
    write_cmd(0x00, (0x80 + 0x40), 0);//2LINE TOP //Display ON + Cursor OFF + Blink OFF
    printString(st2);
  }
}
