/*
  Motion Puttern Builder
  Sample for Arduino
             ESP32-WROOM / ESP32-WROVER

  NIPPON PULSE MOTOR Co., Ltd.
*/

#include "driver/spi_master.h"
#include <Arduino.h>
#include <SPI.h>
#include "MpbSample_data.h"
#include "MpbSample_flow.h"
#include "LCD_control.h"
#include "SW_control.h"
#include "TargetDevice.h"


//----------------------------------------------
//Soft ware virsion ----------------------------
short Version = 0x0001;

//
MPB_flowchart mpb_flow;
Mpb_Samples sample;
LCD_CONTROL lcd;
SW_CONTROL sw;
extern Target_Device pspc4_hw;


//----------------------------------------------
//Debug mode -----------------------------------
#define DEBUG_ON 1
#define DEBUG_OFF 0
byte debugMode = DEBUG_ON;
//
#define DBG(...) debugMode == DEBUG_ON ? Serial.print(__VA_ARGS__): NULL
#define DBG_ln(...) debugMode == DEBUG_ON ? Serial.println(__VA_ARGS__): NULL

extern int ExePoint;//Running number
//??//extern const int TortalPoints;//総実行数


//----------------------------------------------
//SPI transfer ---------------------------------
// Terminal 
#define P_SS     25
#define P_SCK    14
#define P_MOSI   26
#define P_MISO   27
//SPI speed
static const uint32_t P_SPI_CLK = 10000000; // SPI = 10M

//----------------------------------------------
//I2C transfer ---------------------------------
// Terminal 
#define lcd_SDA  21
#define lcd_SCL  22

//----------------------------------------------
//PCD2112 --------------------------------------
// Terminal 
#define ORG      19  //
#define STOP     18  //
#define START     5  //
#define INT      13  //

//----------------------------------------------
//switch ---------------------------------------
// Terminal 
#define UP_SW    36  //
#define DOWN_SW  39  //
#define LEFT_SW  34  //
#define RIGHT_SW 35  //
#define A_SW     32  //
#define B_SW     33  //

//----------------------------------------------
//unused ---------------------------------------
// Terminal 
//#define --       23  // 
//#define --       17  // 
//#define --       16  // 

//----------------------------------------------
//volume ---------------------------------------
// Terminal 
#define VR        4  //

//----------------------------------------------
//SPI send and receive
#define SpiBuffCnts 256     // Max size of SPI transceiver buffer
char SpiBuff[SpiBuffCnts];  // SPI transceiver buffer
int SendCnt = 0;            // Number of transfers

//
int status = 0;

///////////////////////////////////////////////////////////////////////////////
//
unsigned long GetMicroSec()
{
  return(micros());
}

///////////////////////////////////////////////////////////////////////////////
// SPI initial
//
void SPI_init()
{
  SPI.begin(P_SCK, P_MISO, P_MOSI, P_SS);
  SPI.beginTransaction(SPISettings(P_SPI_CLK, MSBFIRST, SPI_MODE0));
  //
  pinMode(P_MISO, INPUT);
  pinMode(P_SS, OUTPUT);
  digitalWrite(P_SS, HIGH);
}


///////////////////////////////////////////////////////////////////////////////
// SPI send and receive
//
void Spi_transfer()
{
  digitalWrite(P_SS, LOW);
  SPI.transfer(SpiBuff, SendCnt); //Receive is stored in [SpiBuff]
  digitalWrite(P_SS, HIGH);
}


///////////////////////////////////////////////////////////////////////////////
void setup()
{
  Serial.begin(115200);
  while (!Serial);

  SPI_init();
  DBG_ln("SPI initial end");

  delay(100);//maybe not necessary

  mpb_flow.begin();
  DBG_ln("LSI(PCL/PCD) initial end");

  lcd.begin();
  DBG_ln("LCD initial end");

  sw.begin();
  DBG_ln("Switch initial end");

  pinMode(P_MISO, INPUT);
  pinMode(INT, INPUT);

  lcd.Display("   MP_Builder", " PUSH <A> start", true);
}

///////////////////////////////////////////////////////////////////////////////
void loop()
{
  if(status == 0)
  {
    if(sw.Check(A_SW) == true)
    {
      status = 1;
      //
      lcd.Display("<<Flow running>>", " PUSH <B> stop", true);
    }
  }
  else if(status == 1)
  {
    if(sample.TortalPoints >= ExePoint)
    {
      mpb_flow.execute();
    }
    else if(sample.TortalPoints < ExePoint)
    {
      DBG_ln("End exe");
      //
      status = 0;
      ExePoint = 0;
      //
      lcd.Display("   MP_Builder", " PUSH <A> start", true);
    }
    //
    if(sw.Check(B_SW) == true)
    {
      status = 0;
      ExePoint = 0;
      //
      pspc4_hw.StopCommand();//Stop-command
      //
      lcd.Display("   MP_Builder", " PUSH <A> start", true);
    }
  }


}
