﻿//////////////////////////////////////////////////////////////////////////////
//
//  Flow chart control source  ( for PCD2112 )
//
//  NIPPON PULSE MOTOR Co., Ltd.
//
//////////////////////////////////////////////////////////////////////////////

using FTD2XX_NET;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace MP_Builder_Sample
{
    public class FlowProcess
    {
        ////////////////////////////////////////////////////////////////////
        //Common processing

        ////////////////////////////////////////////////////////////////////
        [DllImport("winmm.dll", EntryPoint = "timeGetTime")]//高精度な時間待ち
        public static extern uint GetTime();

        ////////////////////////////////////////////////////////////////////
        //Work variables during acceleration/deceleration
        public bool ZeroFlag = false;  // "true" if the addition or subtraction result is zero
        public bool OverFlag = false;  // "true" when the addition/subtraction result is overflow (underflow)
        public byte[] REG = new byte[256];  // Register definition (initialization omitted)

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// register addition
        /// <summary>
        public void REG_Addition(ref byte Target, byte wb1)
        {
          int wr1 = (int)Target + (int)wb1;
          ZeroFlag = false;
          OverFlag = false;
          //overflow
          if(wr1 > 0xFF)
          {
            OverFlag = true;
          }
          //Addition result is zero
          if(wr1 == 0)
          {
            ZeroFlag = true;
          }
          Target = (byte)(wr1 & 0xFF);
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// register subtraction
        /// <summary>
        public void REG_Subtraction(ref byte Target, byte wb1)
        {
          int wr1 = (int)Target - (int)wb1;
          ZeroFlag = false;
          OverFlag = false;
          //underflow
          if(wr1 < 0)
          {
            OverFlag = true;
          }
          //Subtract result is zero
          if(wr1 == 0)
          {
            ZeroFlag = true;
          }
          Target = (byte)(wr1 & 0xFF);
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// WAIT
        /// <summary>
        public void Wait(Form1 fm1, uint WaitTime)
        {
            ulong StartTick = 0;
            uint NowTick = 0;
            ulong ElapsedTick = 0;

            //WAIT processing is performed after the operation of all axes is completed.
            fm1.hFlow.WaitMotionEnd(fm1, 0, fm1.hDataSmp.UseAxis);
            //
            //Start WAIT processing
            if (WaitTime == 0)
            {
                return;
            }
            //
            StartTick = (ulong)GetTime();
            ulong PreTick = StartTick;
            //
            do
            {
                NowTick = GetTime();
                if(PreTick != NowTick)
                {
                    PreTick = NowTick;
                    //
                    if (NowTick > StartTick)
                    {
                        ElapsedTick = (ulong)(NowTick - StartTick);
                    }
                    else
                    {
                        ElapsedTick = (ulong)((0x100000000 -  StartTick) + (ulong)NowTick);
                    }
                }
            } while (ElapsedTick < (ulong)WaitTime) ;
        }

        //
        //PCD2112 unique processing ================================================================
        //
        ////////////////////////////////////////////////////////////////////
        //
        //Number of types of registers required for operation + 1 (start command)
        private const int MaxReg = 9;
        private const int ComNo = 8; // the position of the start command in the array
        //commands
        public byte[] ComVal = new byte[]  { 0x04, //RENV2
                                             0x08, //RENV1
                                             0x0C, //RDS
                                             0x0E, //RUS
                                             0x10, //RDR
                                             0x12, //RUR
                                             0x14, //RMG
                                             0x16, //RDP
                                             0x19, //RFL
                                             0x1B, //RFH1
                                             0x1D, //RFH2
                                             0x1F, //RMV
                                             0x23, //RMD
                                             0x25, //RCOM
                                             0x26, //RSTS
                                             0x2A, //RIST
                                             0x2C, //RIOP
                                             0x2E, //RCUN
                                             0x32, //RDWC
                                             0x36};//RSPD
        private const int RENV2 = 0;
        private const int RENV1 = 1;
        private const int RDS = 2;
        private const int RUS = 3;
        private const int RDR = 4;
        private const int RUR = 5;
        private const int RMG = 6;
        private const int RDP = 7;
        private const int RFL = 8;
        private const int RFH1 = 9;
        private const int RFH2 = 10;
        private const int RMV = 11;
        private const int RMD = 12;
        private const int RCOM = 13;
        private const int RSTS = 14;
        private const int RIST = 15;
        private const int RIOP = 16;
        private const int RCUN = 17;
        private const int RDWC = 18;
        private const int RSPD = 19;
        public int[] RegByte = new int[]  { 4, //RENV2
                                            4, //RENV1
                                            2, //RDS
                                            2, //RUS
                                            2, //RDR
                                            2, //RUR
                                            2, //RMG
                                            3, //RDP
                                            2, //RFL
                                            2, //RFH1
                                            2, //RFH2
                                            4, //RMV
                                            2, //RMD
                                            1, //RCOM
                                            4, //RSTS
                                            2, //RIST
                                            2, //RIOP
                                            4, //RCUN
                                            4, //RDWC
                                            2}; //RSPD
        public int[] RegTble = new int[]  { 11,  // RMV
                                             8,  // RFL
                                             9,  // RFH1
                                             5,  // RUR
                                             4,  // RDR
                                             6,  // RMG
                                             7,  // RDP
                                            12,  // RMD
                                            13}; // RCOM
        //
        public uint c_riop = 0x00000101;//For excitation control

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// MSTS, SSTS, REGD Clear each array
        /// </summary>
        /// <param name="description"></param>
        /// <returns></returns>
        public void InitDatas(Form1 fm1)
        {
            fm1.hFTDI.MSTS[0] = 0;
            fm1.hFTDI.SSTS[0] = 0;
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Generation of general-purpose port data (for excitation control)
        /// </summary>
        /// <param name="com"></param>
        /// <param name="PortSet"></param>
        /// <param name="WkRiop"></param>
        private void ExSetting(uint com, uint PortSet, ref uint WkRiop)
        {
            //
            if (com == 0x02)
            {
                //ReSet(set to ZERO)
                PortSet = (uint)(~PortSet);
                WkRiop = WkRiop & PortSet;
            }
            else if (com == 0x03)
            {
                //Set(set to 1)
                WkRiop = WkRiop | PortSet;
            }
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// LSI initial settings & Excitation control
        /// type = 0 : Initialize RENV1,2 settings and Excitation off.
        ///        1 : Excitation ON only
        ///        2 : Excitation OFF only
        /// </summary>
        /// <param name="fm1"></param>
        /// <param name="type"></param>
        /// <returns></returns>
        public FTDI.FT_STATUS InitPulseGenelator(Form1 fm1, int type)
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;

            //
            fm1.hFTDI.index = 0;
            //
            if(type == 0)
            {
                //RENV1
                fm1.hFTDI.Write_REG(ComVal[RENV1], fm1.hDataSmp.Renv1_ini, RegByte[RENV1]);
                //RENV2
                fm1.hFTDI.Write_REG(ComVal[RENV2], fm1.hDataSmp.Renv2_ini, RegByte[RENV2]);
            }
            //
            uint WkRiop = c_riop;
            int axis = 0;
            //
            if ((type == 0) ||
               (type == 2))
            {
                for (int loopA = 0; loopA < 4; loopA++)//PCD2112 has 4 ports
                {
                    if (fm1.hDataSmp.Port_set[axis, loopA] != 0)
                    {
                        ExSetting(0x02, (byte)(1), ref WkRiop);
                        //PORT
                        fm1.hFTDI.Write_REG(ComVal[RIOP], WkRiop, RegByte[RIOP]);
                    }
                }
            }
            else //if(type == 1)
            {
                for (int loopA = 0; loopA < 4; loopA++)//PCD2112 has 4 ports
                {
                    if (fm1.hDataSmp.Port_set[axis, loopA] != 0)
                    {
                        ExSetting(0x03, (byte)(1), ref WkRiop);
                        //PORT
                        fm1.hFTDI.Write_REG(ComVal[RIOP], WkRiop, RegByte[RIOP]);
                    }
                }
            }
            //
            InitDatas(fm1);
            //
            ftStatus = fm1.hFTDI.SendUsb(fm1.hFTDI.WriteBuffer, ref fm1.hFTDI.index);
            //
            return (ftStatus);
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Check the operation status of the specified axis
        ///   CheckSts：Comparison value with Status
        ///             Wait until Status is less than or equal to this value
        ///             
        ///             Return value "Status" of "hFTDI.WaitMsts" is as follows
        ///                 Status＝0：Stopped (the next operation can be written)
        ///                 Status＝1：In operation, pre-registration is free (the next operation can be written)
        ///                 Status＝2：undefined
        ///                 Status＝3：pre-register full (cannot write next operation)
        /// </summary>
        /// <param name="CheckAxis"></param>
        public void WaitMotionEnd(Form1 fm1, int CheckSts, int CheckAxis)
        {
            int EvCnt = 0;
            int Status = 3;
            while (Status > CheckSts)
            {
                EvCnt++;
                if (EvCnt > 100)
                {
                    EvCnt = 0;
                    System.Windows.Forms.Application.DoEvents();
                }
                //
                Status = fm1.hFTDI.WaitMsts(fm1, fm1.hDataSmp.AxisConts, CheckAxis);
                //
                if (fm1.hFTDI.FlowData_stop == true)//Stop processing was performed on the Form1 side
                {
                    //Write the STOP command to all axes when interrupting
                    fm1.hFTDI.index = 0;
                    fm1.hFTDI.Write_COM((byte)0x31);
                    fm1.hFTDI.SendUsb(fm1.hFTDI.WriteBuffer, ref fm1.hFTDI.index);
                    //
                    fm1.hFTDI.FlowData_on = true;
                }
            }
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        ///  Execute positioning movement
        ///  (LSI access processing)
        /// <summary>
        public void MotioneProcess(Form1 fm1, int DataNum)
        {
            //For grouping the same start command
            byte[] StCom = new byte[] {0x40, 0x41, 0x42, 0x43, 0x44};
            byte[] StAx = new byte[] {0, 0, 0, 0, 0, 0 };
            int UseAxis = 0;//Information about the axis to be executed
            //
            fm1.hFTDI.index = 0;
            //
            if(fm1.hDataSmp.RegDat[DataNum, ComNo] != 0)
            {
                //Axes with no start command specified are not subject to execution
                //
                //Create register write data
                for(int loopB = 0; loopB < ComNo; loopB++)
                {
                    fm1.hFTDI.Write_REG(ComVal[RegTble[loopB]], fm1.hDataSmp.RegDat[DataNum, loopB], RegByte[RegTble[loopB]]);
                }
                //
                //group the same start commands together
                //StAx[] holds axis information
                switch(fm1.hDataSmp.RegDat[DataNum, ComNo])
                {
                    case 0x40://STAFL
                        StAx[0] = (byte)0x01;
                        break;
                    case 0x41://STAFH1
                        StAx[1] = (byte)0x01;
                        break;
                    case 0x42://STAFH2
                        StAx[2] = (byte)0x01;
                        break;
                    case 0x43://STAUD1
                        StAx[3] = (byte)0x01;
                        break;
                    case 0x44://STAUD2
                        StAx[4] = (byte)0x01;
                        break;
                    case 0xFF://FH1 -> FL 
                        StAx[5] = (byte)0x01;
                        break;
                }
                //
                UseAxis = 0x01;
            }
            //
            //Create write data for start command
            for(int loopB = 0; loopB < 6; loopB++)
            {
                if(StAx[loopB] != 0)
                {
                    if (loopB < 5)
                    {
                        fm1.hFTDI.Write_COM(StCom[loopB]);
                    }
                    else
                    {
                        //Switch to "Decelerate to FL at FH start" operation
                        fm1.hFTDI.Write_COM(StCom[1]);
                        //
                        if ((fm1.hDataSmp.Renv1_ini & 0x00800000) != 0)
                        {
                            fm1.hFTDI.Write_COM((byte)(0x00));//NOP
                            fm1.hFTDI.Write_COM((byte)(0x00));//NOP
                            fm1.hFTDI.Write_COM((byte)(0x00));//NOP
                            fm1.hFTDI.Write_COM((byte)(0x00));//NOP
                        }
                        //
                        fm1.hFTDI.Write_COM((byte)0x63);//Change to "Decelerate and FL speed"
                    }
                }
            }
            //Wait until the LSI is ready to write the next command
            WaitMotionEnd(fm1, 1, UseAxis);
            //
            //Send command to USB
            if(fm1.hFTDI.FlowData_stop == false)
            {
                fm1.hFTDI.SendUsb(fm1.hFTDI.WriteBuffer, ref fm1.hFTDI.index);
            }
        }
    }
}
