/*
  Motion Puttern Builder
  Sample for Arduino
             ESP32-WROOM / ESP32-WROVER

  Target is PCD2112

  NIPPON PULSE MOTOR Co., Ltd.
*/

#include "MpbSample_data.h"
#include "TargetDevice.h"

extern Mpb_Samples sample;
unsigned long GetMicroSec();

//----------------------------------------------
//Flow controll
//Variables used in WAIT
//Manage time in microseconds
bool WaitEn = false;
unsigned long TsTime = 0;//Holds the initial time used in WAIT
unsigned long WaitT;//wait time
unsigned long NwTime;
unsigned long CpTime;//For comparison
//Used in register operations
unsigned short REG[WorkBuffSize];//Register on Flow
bool ZeroFlag = false;
bool OverFlag = false;


//----------------------------------------------
//Targer device controll
char PreRegStatus; //PreRegister status 0:Not Busy   1:Busy bat Not use PreRegister  2:PreRegister is full
int ExePoint = 0;//Execution point

//----------------------------------------------
//SPI send and receive
#define SpiBuffCnts 256     // Max size of SPI transceiver buffer
extern char SpiBuff[SpiBuffCnts];  // SPI transceiver buffer
extern int SendCnt;                // Number of transfers

void Spi_transfer();


//----------------------------------------------
//Targer device main-status and port
long Msts[1] = {0};
char Port[1] = {0};

/*//Register address
int RegAdd[] = {0x1F, // RMV
                0x19, // RFL
                0x1B, // RFH1
                0x12, // RUR
                0x10, // RDR
                0x16, // RDP
                0x23  // RMD
               };*/

//Register size
int RegSize[] = {4, // RMV
                 2, // RFL
                 2, // RFH1
                 2, // RUR
                 2, // RDR
                 3, // RDP
                 2  // RMD
                };


//----------------------------------------------
//UNION
union ShortChar
{
  char CH[2];
  short SH;
};
//
union LongChar
{
  char CH[4];
  long LH;
};


///////////////////////////////////////////////////////////////////////////////
// Read status from Targer device
//
void Target_Device::Read_Status()
{
  int DumyCnt = 4;
  SendCnt = 0;
  //
  SpiBuff[SendCnt++] = (sample.Ds << 4) | 0x03; // read
  SpiBuff[SendCnt++] = 0x26; // status-address
  //
  for (int loopA = 0; loopA < DumyCnt; loopA++)
  {
    SpiBuff[SendCnt++] = 0;//Dumy
  }
  //
  Spi_transfer();
  //
  int GetPoint = 2;
  LongChar lc;//union
  //
  lc.CH[0] = SpiBuff[GetPoint++];
  lc.CH[1] = SpiBuff[GetPoint++];
  lc.CH[2] = SpiBuff[GetPoint++];
  lc.CH[3] = SpiBuff[GetPoint++];
  Msts[0] = lc.LH;
}


///////////////////////////////////////////////////////////////////////////////
// Read port from Targer device
//
void Target_Device::Read_Port()
{
  int DumyCnt = 2;
  SendCnt = 0;
  //
  SpiBuff[SendCnt++] = (sample.Ds << 4) | 0x03; // read
  SpiBuff[SendCnt++] = 0x2C; // port-address
  //
  for (int loopA = 0; loopA < DumyCnt; loopA++)
  {
    SpiBuff[SendCnt++] = 0;//Dumy
  }
  //
  Spi_transfer();
  ShortChar sc;//union
  //
  int GetPoint = 2;
  //
  sc.CH[0] = SpiBuff[GetPoint++]; // port-data
  sc.CH[1] = SpiBuff[GetPoint++]; // port-io
  Port[0] =  sc.SH;
}


///////////////////////////////////////////////////////////////////////////////
// Check pre-register status
//
// return 0: Device is not busy
//        1: Device is Busy
//
int Target_Device::CheckPreReg()
{
  int ret = 0;
  //
  int loopA = 0;//X-axis only
  if((Msts[loopA] & 0x0F000000) != 0x00000000)  // Device is Busy
  {
    ret = 1;
  }
  //
  return(ret);
}


///////////////////////////////////////////////////////////////////////////////
// Operating status check
//
int Target_Device::CheckMotionStatus()
{
    Read_Status();
    PreRegStatus = CheckPreReg();
    //
    if (PreRegStatus == 0)
    {
        //all actions completed
        return(1);
    }
    else
    {
        return(0);
    }
 }


///////////////////////////////////////////////////////////////////////////////
// 
//
int Target_Device::MotioneProcess(int num)
{
  int ret = 0;

  Read_Status();
  PreRegStatus = CheckPreReg();

  //
  if(PreRegStatus == 0) //Device is not busy.
  {
    LongChar lc;//union

    //
    //Send Register write command.
    for(int loopA = 0; loopA < RegCnt; loopA++)
    {
      //
      //Register write command set to buffer.
      int loopB = 0;
      //
      if(sample.RegDat[num][loopB][7] != 0)  // This axis is use
      {
        SendCnt = 0;
        //
        SpiBuff[SendCnt++] = (sample.Ds << 4) | 0x02; // write
        SpiBuff[SendCnt++] = sample.RegWt[loopA]; // Write address
        //
        lc.LH = sample.RegDat[num][loopB][loopA];
        //
        if(RegSize[num] > 3)
        {
          SpiBuff[SendCnt++] = lc.CH[3];
        }
        if(RegSize[num] > 2)
        {
          SpiBuff[SendCnt++] = lc.CH[2];
        }
        if(RegSize[num] > 1)
        {
          SpiBuff[SendCnt++] = lc.CH[1];
        }
        SpiBuff[SendCnt++] = lc.CH[0];
        //
        Spi_transfer();
      }
    }//Send Register write command.
    //
    //Start command set to buffer
    int loopA = 0;//X-axis only
    SendCnt = 0;
    SpiBuff[SendCnt++] = (sample.Ds << 4) | 0x02; // write
    SpiBuff[SendCnt++] = 0x25; // command address
    if(sample.RegDat[num][loopA][7] = 0x42)//FH2-start
    {
      //"Deceleration stop at FH start" operation
      SpiBuff[SendCnt++] = 0x41;//FH1スタート
      Spi_transfer();
      //
      if((sample.Renv1_ini[0] & 0x00800000) != 0)
      {
        for(int loopC = 0; loopC < 4; loopC++)
        {
          SendCnt = 0;
          SpiBuff[SendCnt++] = (sample.Ds << 4) | 0x02; // write
          SpiBuff[SendCnt++] = 0x25; // command address
          SpiBuff[SendCnt++] = 0x00; // NOP
          Spi_transfer();
        }
      }
      //
      SendCnt = 0;
      SpiBuff[SendCnt++] = (sample.Ds << 4) | 0x02; // write
      SpiBuff[SendCnt++] = 0x25; // command address
      SpiBuff[SendCnt++] = 0x63; // Decelerate and change to FL speed
      Spi_transfer();
    }
    else
    {
      SpiBuff[SendCnt++] = sample.RegDat[num][loopA][7];
    }
    Spi_transfer();
    
    //
    ret = 1;
  }
  //
  //ret = 1; //<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Absolutely Delete !!!!!!!!!!!!!!!!
  return(ret);
}


///////////////////////////////////////////////////////////////////////////////
// Write PCD2112-register
//
void Target_Device::WriteReg_hw_ini(char com, short *data)
{
  LongChar lc;//union

  //
  SendCnt = 0;
  SpiBuff[SendCnt++] = (sample.Ds << 4) | 0x02; // write
  SpiBuff[SendCnt++] = sample.RegWt[com]; // Write command
  //
  lc.LH = *(data);
  if(RegSize[com] > 3)
  {
    SpiBuff[SendCnt++] = lc.CH[3];
  }
  if(RegSize[com] > 2)
  {
    SpiBuff[SendCnt++] = lc.CH[2];
  }
  if(RegSize[com] > 1)
  {
    SpiBuff[SendCnt++] = lc.CH[1];
  }
  SpiBuff[SendCnt++] = lc.CH[0];
  //
  Spi_transfer();
}


///////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////
// WAIT operation
//
// If any axis are in motion, all axis are stopped and called.
///////////////////////////////////////////////////////////////////////////////////////
int Target_Device::WaitExe(short WtTime)
{
  int ret = 0;//Default is to not proceed

  if (WaitEn == false) //Not in WAIT processing
  {
    Read_Status();
    PreRegStatus = CheckPreReg(0x0F);
    if (PreRegStatus == 0)
    {
      //
      WaitT = WtTime * 1000;//to microseconds
      WaitEn = true;
      //
      TsTime = GetMicroSec();//time at start
    }
  }
  else//WAIT in operation
  {
    NwTime = GetMicroSec();
    if (TsTime > NwTime) //timer overflowed(70 minutes passed)
    {
      NwTime = (0xFFFFFFFF - TsTime)  + NwTime + 1;
    }
    else
    {
      NwTime = NwTime - TsTime;
    }
    //WAIT process end
    if (NwTime >= WaitT) //WAIT end
    {
      //
      WaitEn = false;
      //
      ret = 1;//proceed
    }
  }
  //
  return (ret);
}


///////////////////////////////////////////////////////////////////////////////////////
// Register Subtraction operation
//
//  Comm  : Operation type
//  RegNo : Register number (storage destination)
//  dat   : Data or Data source register number
///////////////////////////////////////////////////////////////////////////////////////
void CheckStattValue(unsigned long RegFlag)
{
  //
  if((RegFlag & 0x0000FFFF) == 0)
  {
    ZeroFlag = true;
  }
  else
  {
    ZeroFlag = false;
  }
  //
  if((RegFlag & 0xFFFF0000) != 0)
  {
    OverFlag = true;
  }
  else
  {
    OverFlag = false;
  }
}


///////////////////////////////////////////////////////////////////////////////////////
// Register Addition operation
//
//  Comm  : Operation type
//  RegNo : Register number (storage destination)
//  dat   : Data or Data source register number
///////////////////////////////////////////////////////////////////////////////////////
unsigned short Target_Device::REG_Addition(unsigned short Reg, int dat)
{
  unsigned long RegFlag;
  //
  RegFlag = (unsigned long)Reg + (unsigned long)dat;
  CheckStattValue(RegFlag);
  //
  return((unsigned short)(RegFlag & 0x0000FFFF));
}


///////////////////////////////////////////////////////////////////////////////////////
// Register Subtraction operation
//
//  Comm  : Operation type
//  RegNo : Register number (storage destination)
//  dat   : Data or Data source register number
///////////////////////////////////////////////////////////////////////////////////////
unsigned short Target_Device::REG_Subtraction(unsigned short Reg, int dat)
{
  unsigned long RegFlag;
  //
  RegFlag = (unsigned long)Reg - (unsigned long)dat;
  CheckStattValue(RegFlag);
  //
  return((unsigned short)(RegFlag & 0x0000FFFF));
}


///////////////////////////////////////////////////////////////////////////////////////
// Stop command transfer operation
//
///////////////////////////////////////////////////////////////////////////////////////
void Target_Device::StopCommand()
{
  //
  SendCnt = 0;
  SpiBuff[SendCnt++] = (sample.Ds << 4) | 0x02; // write
  SpiBuff[SendCnt++] = 0x25; // command address
  SpiBuff[SendCnt++] = 0x31; // Stop
  //
  Spi_transfer();
}
