﻿//////////////////////////////////////////////////////////////////////////////
//
//  FTDI access source  ( for PCD46x1 )
//
//  NIPPON PULSE MOTOR Co., Ltd.
//
//////////////////////////////////////////////////////////////////////////////

using System;
using System.Runtime.InteropServices;
using FTD2XX_NET;
using System.Threading;
using System.Diagnostics;
using System.Linq;

namespace MP_Builder_Sample
{
    public partial class clsFTDI
    {
        ////////////////////////////////////////////////////////////////////
        /// 
        public FTDI ftHandle = new FTDI();
        //
        public byte Ds = 0;//Device number is changed in software
        //keep status
        public uint[] MSTS = new uint[4];
        public uint[] SSTS = new uint[4];
        //
        //Maximum definition
        private const int MAX_TAG = 65536;
        private const int MAX_REG = 256;
        private const int MAX_BUFF = 4096;
        //
        public int index = 0;
        public byte[] WriteBuffer = new byte[MAX_BUFF];
        public int indexS = 0;//for status read
        public byte[] WriteBufferS = new byte[MAX_BUFF];//for status read
        public byte[] ReadBuffer = new byte[MAX_BUFF];
        public bool FlowData_on = true;//active flag
        public bool FlowData_stop = false;//Suspend flag = off

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// use as union
        /// </summary>
        [StructLayout(LayoutKind.Explicit)]
        public struct TBytes
        {
            [FieldOffset(0)]
            public byte Byte0;

            [FieldOffset(1)]
            public byte Byte1;

            [FieldOffset(2)]
            public byte Byte2;

            [FieldOffset(3)]
            public byte Byte3;

            [FieldOffset(0)]
            public uint Word;
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// FTDI connection processing
        /// </summary>
        /// <param name="description"></param>
        /// <returns></returns>
        public FTDI.FT_STATUS LSI_Open(Form1 fm1, string description)
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;

            ftStatus = ftHandle.OpenByDescription(description);

            if (!ftHandle.IsOpen)
            {   // error
                Debug.WriteLine("Error: FTDI.penByDescription! [ " + ftStatus.ToString() + " ]");
                return ftStatus;
            }

            ////////////////////////////////////////////////////////////////////
            // FTDI.ResetDevice
            ftStatus = ftHandle.ResetDevice();

            if (ftStatus != FTDI.FT_STATUS.FT_OK)
            {   // error
                Debug.WriteLine("Error: FTDI.ResetDevice! [ " + ftStatus.ToString() + " ]");
                ftHandle.Close();
                return ftStatus;
            }

            Thread.Sleep(100);     //Wair fot reset process

            ////////////////////////////////////////////////////////////////////
            // FTDI.SetLatency (1ms)
            ftStatus = ftHandle.SetLatency(1);

            if (ftStatus != FTDI.FT_STATUS.FT_OK)
            {   // error
                Debug.WriteLine("Error: FTDI.SetLatency! [ " + ftStatus.ToString() + " ]");
                ftHandle.Close();
                return ftStatus;
            }

            ////////////////////////////////////////////////////////////////////
            // FTDI.SetTimeouts
            uint ReadTimeout = 1000;    // 1000ms
            uint WriteTimeout = 1000;   // 1000ms

            ftStatus = ftHandle.SetTimeouts(ReadTimeout, WriteTimeout);

            if (ftStatus != FTDI.FT_STATUS.FT_OK)
            {   // error
                Debug.WriteLine("Error: FTDI.SetTimeouts! [ " + ftStatus.ToString() + " ]");
                ftHandle.Close();
                return ftStatus;
            }

            ////////////////////////////////////////////////////////////////////
            // FTDI.SetBitMode (MPSSE Mode)
            byte Mask = 0;
            byte BitMode = FTDI.FT_BIT_MODES.FT_BIT_MODE_MPSSE;

            ftStatus = ftHandle.SetBitMode(Mask, BitMode);

            if (ftStatus != FTDI.FT_STATUS.FT_OK)
            {   // error
                Debug.WriteLine("Error: FTDI.SetBitMode! [ " + ftStatus.ToString() + " ]");
                ftHandle.Close();
                return ftStatus;
            }
            //
            ////////////////////////////////////////////////////////////////////
            // FTDI.Write (SPI=10MHz)
            byte[] dataBuffer = new byte[] { 0x8D, // Disable 3 Phase Data Clocking (Waveforms other than I2C)
                                             0x86, 0x02, 0x00 };    // SPI=10MHz
            int numBytesToWrite = 4;
            uint numBytesWritten = 0;

            ftStatus = ftHandle.Write(dataBuffer, numBytesToWrite, ref numBytesWritten);

            if (ftStatus != FTDI.FT_STATUS.FT_OK)
            {   // error
                Debug.WriteLine("Error: FTDI.Write! [ " + ftStatus.ToString() + " ]");
                return ftStatus;
            }

            ////////////////////////////////////////////////////////////////////
            // FTDI.Write (Set initial port value)
            dataBuffer[0] = 0x80;
            dataBuffer[1] = 0x18;//RST=High, CS=High
            dataBuffer[2] = 0x1B;
            numBytesToWrite = 3;

            ftStatus = ftHandle.Write(dataBuffer, numBytesToWrite, ref numBytesWritten);

            if (ftStatus != FTDI.FT_STATUS.FT_OK)
            {   // error
                Debug.WriteLine("Error: FTDI.Write! [ " + ftStatus.ToString() + " ]");
                return ftStatus;
            }

            ////////////////////////////////////////////////////////////////////
            // FTDI.Write (12MHz -> 60MHz)  SPI is 10MHz
            dataBuffer[0] = 0x8A;// 60MHz
            numBytesToWrite = 1;
            numBytesWritten = 0;

            ftStatus = ftHandle.Write(dataBuffer, numBytesToWrite, ref numBytesWritten);

            if (ftStatus != FTDI.FT_STATUS.FT_OK)
            {   // error
                Debug.WriteLine("Error: FTDI.Write! [ " + ftStatus.ToString() + " ]");
                return ftStatus;
            }

            ////////////////////////////////////////////////////////////////////
            // FTDI.Purge
            //  Clear transmit and receive buffers
            ftStatus = ftHandle.Purge(FTDI.FT_PURGE.FT_PURGE_RX | FTDI.FT_PURGE.FT_PURGE_TX);

            if (ftStatus != FTDI.FT_STATUS.FT_OK)
            {   // error
                Debug.WriteLine("Error: FTDI.Purge! [ " + ftStatus.ToString() + " ]");
                ftHandle.Close();
                return ftStatus;
            }

            ////////////////////////////////////////////////////////////////////
            return ftStatus;
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public FTDI.FT_STATUS LSI_Close()
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;

            ////////////////////////////////////////////////////////////////////
            // FTDI.Close
            ftStatus = ftHandle.Close();

            if (ftHandle.IsOpen)
            {   // error
                Debug.WriteLine("Error: FTDI.Close! [" + ftStatus.ToString() + "]");
                return ftStatus;
            }

            return ftStatus;
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// FTDI Write
        /// </summary>
        /// <param name="dataBuffer"></param>
        /// <param name="numBytesToWrite"></param>
        /// <returns></returns>
        public FTDI.FT_STATUS LSI_Write(byte[] dataBuffer, int numBytesToWrite)
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;

            ////////////////////////////////////////////////////////////////////
            // FTDI.Write
            uint numBytesWritten = 0;

            ftStatus = ftHandle.Write(dataBuffer, numBytesToWrite, ref numBytesWritten);

            ////////////////////////////////////////////////////////////////////
            // Error Check
            if (ftStatus != FTDI.FT_STATUS.FT_OK)
            {   // error
                Debug.WriteLine("Error: FTDI.Write! [ " + ftStatus.ToString() + " ]");
                return ftStatus;
            }

            if (numBytesToWrite != numBytesWritten)
            {   // error
                Debug.WriteLine(
                    "Error: FTDI.Write! [ numBytesToWrite: " + numBytesToWrite.ToString() +
                    " /= numBytesWritten: " + numBytesWritten.ToString() + " ]");
                return FTDI.FT_STATUS.FT_OTHER_ERROR;
            }

            ////////////////////////////////////////////////////////////////////
            return ftStatus;
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Garbage may remain after forced termination.
        /// This function eliminates garbage.
        /// </summary>
        /// <param name="dataBuffer"></param>
        /// <returns></returns>
        public uint LSI_CheckRead(ref byte[] dataBuffer)
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;
            uint RxQueue = 0;
            uint Dumy = 0;

            ftStatus = ftHandle.GetRxBytesAvailable(ref RxQueue);
            //
            if ((ftStatus == FTDI.FT_STATUS.FT_OK) && (RxQueue > 0))
            {
                ftStatus = ftHandle.Read(dataBuffer, RxQueue, ref Dumy);
                RxQueue = Dumy;
            }
            //
            return(RxQueue);
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Pick up register data from USB read buffer "ReadBuffer"
        ///   TargetReg ： Target : target number
        ///   ofset ： offset to start of data
        /// </summary>
        /// <param name="description"></param>
        /// <returns></returns>
        public void PickUpRegdat(Form1 fm1, int TargetReg, int ofset)
        {
            var bytes = new TBytes();
            //
            bytes.Byte3 = 0;//Data 31-24
            //
            int sft = 16; // for RIDC
            //
            if (TargetReg == 14) // RSPD
            {
                sft = 8;
            }
            //
            for (int loops = 0, of1; loops < fm1.hDataSmp.AxisConts; loops++)
            {
                of1 = loops * 3 + ofset;
                bytes.Byte0 = (byte)(ReadBuffer[of1++]);//Data 7-0
                bytes.Byte1 = (byte)(ReadBuffer[of1++]);//Data 15-8
                bytes.Byte2 = (byte)(ReadBuffer[of1++]);//Data 23-16
                //
                uint wr1 = (uint)(bytes.Word);
                if ((TargetReg >= 13) && (TargetReg <= 14))
                {
                    wr1 = wr1 >> sft;
                }
                //
                //Mask
                fm1.hFlow.REGD[loops, TargetReg] = (uint)(wr1 & fm1.hFlow.RegMask[TargetReg]);
            }
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Pick up port data from USB read buffer "ReadBuffer"
        ///   TargetReg ： Target : target number
        ///   ofset ： offset to start of data
        /// </summary>
        /// <param name="description"></param>
        /// <returns></returns>
        public void PickUpPort(Form1 fm1, int ofset)
        {
            int TargetReg = FlowProcess.RIOP;
            //
            for (int loops = 0, ofs = 0; loops < fm1.hDataSmp.AxisConts; loops++)
            {
                ofs = loops + ofset;
                fm1.hFlow.REGD[loops, TargetReg] = (uint)(ReadBuffer[ofs]);
            }
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// 
        /// </summary>
        /// <param name="fm1"></param>
        public void Read_Port(Form1 fm1)
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;
            index = 0;
            //
            byte jj = 0;
            switch (fm1.hDataSmp.AxisConts)
            {
                case 2:
                    jj = 0x23;//2軸
                    break;
                case 4:
                    jj = 0x2F;//4軸
                    break;
                default:
                    jj = 0x21;//1軸
                    break;
            }
            //
            //CS=Low
            WriteBuffer[index++] = 0x80;
            WriteBuffer[index++] = 0x00;
            WriteBuffer[index++] = 0x0B;
            //
            WriteBuffer[index++] = 0x35;//SPI(Out on negative edge, in on negative edge)
            WriteBuffer[index++] = (byte)(fm1.hDataSmp.AxisConts);//Length L (MaxAxis+1)byte
            WriteBuffer[index++] = 0x00;//Length H
            WriteBuffer[index++] = (byte)(jj | (Ds << 6));// 0x2n General purpose port lead, All axes
            for (int loops = 0; loops < fm1.hDataSmp.AxisConts; loops++)
            {
                WriteBuffer[index++] = 0x00;//axis dumy
            }
            //CS=High
            WriteBuffer[index++] = 0x80;
            WriteBuffer[index++] = 0x08;
            WriteBuffer[index++] = 0x0B;
            //
            WriteBuffer[index++] = 0x87;
            //
            ftStatus = fm1.hFTDI.LSI_Write(WriteBuffer, index);
            //
            if (ftStatus == FTDI.FT_STATUS.FT_OK)
            {
                uint numBytesRead = 5;
                ftStatus = LSI_Read_first(ref ReadBuffer, ref numBytesRead);
            }

            //
            PickUpPort(fm1, 1);
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        ///   TargetReg  ： Register to read
        /// </summary>
        /// <param name="description"></param>
        /// <returns></returns>
        public void Read_REG(Form1 fm1, int TargetReg)
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;
            index = 0;
            //
            byte jj = 0;
            switch (fm1.hDataSmp.AxisConts)
            {
                case 2:
                    jj = 0x13;//2-axis
                    break;
                case 4:
                    jj = 0x1F;//4-axis
                    break;
                default:
                    jj = 0x11;//1-axis
                    break;
            }
            //
            int comm = fm1.hFlow.ComVal[TargetReg];
            //
            //CS=Low
            WriteBuffer[index++] = 0x80;
            WriteBuffer[index++] = 0x00;
            WriteBuffer[index++] = 0x0B;
            //
            WriteBuffer[index++] = 0x35;//SPI(Out on negative edge, in on negative edge)
            WriteBuffer[index++] = (byte)(fm1.hDataSmp.AxisConts * 3 + 1);//Length L (MaxAxis*3+2)byte
            WriteBuffer[index++] = 0x00;//Length H
            WriteBuffer[index++] = (byte)(jj | (Ds << 6));// 0x1n General purpose lead, All axes
            WriteBuffer[index++] = (byte)(0x80 | comm);
            //
            for (int loops = 0; loops < fm1.hDataSmp.AxisConts; loops++)
            {
                WriteBuffer[index++] = 0x00;//axis dumy
                WriteBuffer[index++] = 0x00;
                WriteBuffer[index++] = 0x00;
            }
            //CS=High
            WriteBuffer[index++] = 0x80;
            WriteBuffer[index++] = 0x08;
            WriteBuffer[index++] = 0x0B;
            //
            WriteBuffer[index++] = 0x87;
            //
            ftStatus = fm1.hFTDI.LSI_Write(WriteBuffer, index);
            //
            if (ftStatus == FTDI.FT_STATUS.FT_OK)
            {
                uint numBytesRead = 0;
                ftStatus = LSI_Read_first(ref ReadBuffer, ref numBytesRead);
            }

            //
            PickUpRegdat(fm1, TargetReg, 2);
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// FTDI Read(Hight speed)
        /// </summary>
        /// <param name="dataBuffer"></param>
        /// <param name="numBytesRead"></param>
        /// <returns></returns>
        public FTDI.FT_STATUS LSI_Read_first(ref byte[] dataBuffer, ref uint numBytesRead)
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;

            ////////////////////////////////////////////////////////////////////
            // FTDI.Read

            int bad = 0;
            uint RxQueue = 0;
            do
            {
                ftStatus = ftHandle.GetRxBytesAvailable(ref RxQueue);
                bad++;
                if (bad > 5000)
                {
                    ftStatus = FTDI.FT_STATUS.FT_EEPROM_READ_FAILED;
                    break;
                }

            } while (RxQueue < numBytesRead);

            uint numBytesToRead = 0;
            do
            {
                ftStatus = ftHandle.Read(dataBuffer, numBytesRead, ref numBytesToRead);

                if (ftStatus != FTDI.FT_STATUS.FT_OK)
                {   // error
                    Debug.WriteLine("Error: FTDI.Read! [" + ftStatus.ToString() + "]");
                    return ftStatus;
                }
                //
                numBytesRead = numBytesRead - numBytesToRead;
            }
            while (numBytesRead > 0);

            return ftStatus;
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Set buffer the command Note that only one axis is processed
        ///   comm ： register write command
        ///   RegD ： write data
        /// </summary>
        /// <returns></returns>
        public void Write_REG(byte Jsc, byte comm, uint RegD, int mask)
        {
            ////////////////////////////////////////////////////////////////////
            var bytes = new TBytes();

            bytes.Word = (uint)(RegD & mask); //set to union

            //CS=Low
            WriteBuffer[index++] = 0x80;
            WriteBuffer[index++] = 0x00;
            WriteBuffer[index++] = 0x0B;
            //
            WriteBuffer[index++] = 0x11;//SPI(Clock Data Bytes Out on +ve Clock Edge MSB First (no Read))
            WriteBuffer[index++] = 0x04;//Length L 5byte
            WriteBuffer[index++] = 0x00;//Length H
            WriteBuffer[index++] = (byte)((Ds << 6) | 0x00 | Jsc); // DS="00", general purpose light, axis designation
            WriteBuffer[index++] = (byte)(0x80 | comm);
            WriteBuffer[index++] = bytes.Byte0;//Data 7-0
            WriteBuffer[index++] = bytes.Byte1;//Data 15-8
            WriteBuffer[index++] = bytes.Byte2;//Data 23-16
            //PileUpBuffer[index++] = bytes.Byte3;//Data 31-24
            //CS=High
            WriteBuffer[index++] = 0x80;
            WriteBuffer[index++] = 0x08;
            WriteBuffer[index++] = 0x0B;
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Set buffer the command Note that only one axis is processed
        ///   comm ： command
        /// </summary>
        /// <returns></returns>
        public void Write_COM(byte Jsc, byte comm)
        {
            ////////////////////////////////////////////////////////////////////
            //CS=Low
            WriteBuffer[index++] = 0x80;
            WriteBuffer[index++] = 0x00;
            WriteBuffer[index++] = 0x0B;
            //
            WriteBuffer[index++] = 0x11;//SPI(Clock Data Bytes Out on +ve Clock Edge MSB First (no Read))
            WriteBuffer[index++] = 0x01;//Length L 2byte
            WriteBuffer[index++] = 0x00;//Length H
            WriteBuffer[index++] = (byte)((Ds << 6) | 0x00 | Jsc); // DS="00", general purpose light, axis designation
            WriteBuffer[index++] = comm;
            //CS=High
            WriteBuffer[index++] = 0x80;
            WriteBuffer[index++] = 0x08;
            WriteBuffer[index++] = 0x0B;
       }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Send seted data via USB
        /// </summary>
        /// <returns></returns>
        public FTDI.FT_STATUS SendUsb(byte[] WriteBuffer, ref int index)
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;
            //
            if (index > 0)
            {
                ftStatus = LSI_Write(WriteBuffer, index);
                index = 0;
            }
            //
            return (ftStatus);
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Set b4ffer teh status read command
        /// </summary>
        public void Read_Status(int AxisConts)
        {
            //
            byte jj = 0;
            switch (AxisConts)
            {
                case 2:
                    jj = 0x33;//2-axis
                    break;
                case 4:
                    jj = 0x3F;//4-axis
                    break;
                default:
                    jj = 0x31;//1-axis
                    break;
            }
            //
            indexS = 0;
            //
            //  status ------------------------------------------------
            //CS=Low
            WriteBufferS[indexS++] = 0x80;
            WriteBufferS[indexS++] = 0x00;
            WriteBufferS[indexS++] = 0x0B;
            //
            WriteBufferS[indexS++] = 0x35;//SPI(Out on negative edge, in on negative edge)
            WriteBufferS[indexS++] = (byte)(AxisConts);//Length L (MaxAxis+1)byte
            WriteBufferS[indexS++] = 0x00;//Length H
            WriteBufferS[indexS++] = (byte)(jj | (Ds << 6));// Ds, 0x3n MainStatus read, All axis
            for (int loops = 0; loops < AxisConts; loops++)
            {
                WriteBufferS[indexS++] = 0x00;//axis dumy
            }
            //CS=High
            WriteBufferS[indexS++] = 0x80;
            WriteBufferS[indexS++] = 0x08;
            WriteBufferS[indexS++] = 0x0B;
            //
            WriteBufferS[indexS++] = 0x87;
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Pick up status and sub-status from USB read buffer "ReadBuffer"
        ///   ofset ： offset to the beginning of the data (not used in this source, just match the other arguments)
        ///   MSTS ： Main status read result
        ///   SSTS  ： Substatus read result
        /// </summary>
        /// <returns></returns>
        public void PickUpStatus(Form1 fm1, int ofset)
        {
            //
            for (int loops = 0; loops < fm1.hDataSmp.AxisConts; loops++)
            {
                MSTS[loops] = (uint)(ReadBuffer[loops + ofset]);
            }
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Wait for operation completion
        /// Do not move to the next operation until all of the specified axes are completed or the pre-registration is free
        ///
        ///  AxisConts : Number of LSI axes
        ///  UseAxis   : Specify axis to wait for status
        ///
        ///  return    : 0＝All specified axes are stopped
        ///            : 1＝All specified axes are in motion but pre-registration is empty or stopped
        ///            : 3＝One of the specified axes is in motion and pre-registration is not free
        /// </summary>
        public int WaitMsts(Form1 fm1, int AxisConts, int UseAxis)
        {
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;
            int[] cksts = new int[] {3, 3, 3, 3};
            //
            indexS = 0;
            Read_Status(AxisConts);
            //
            ftStatus = SendUsb(WriteBufferS, ref indexS);
            //
            if (ftStatus == FTDI.FT_STATUS.FT_OK)
            {
                uint numBytesRead = (uint)(AxisConts + 1);
                ftStatus = LSI_Read_first(ref ReadBuffer, ref numBytesRead);
                //
                if (ftStatus == FTDI.FT_STATUS.FT_OK)
                {
                    PickUpStatus(fm1, 1);
                }
                else
                {
                    //Forced termination because it could not be read
                    MSTS[0] = 0;
                }
            }
            //
            for (int loopA = 0, ax = 1; loopA < 4; loopA++)
            {
                if ((UseAxis & ax) == ax)
                {
                    if ((MSTS[loopA] & 0x00000008) == 0)
                    {
                        //stopped
                        cksts[loopA] = 0;
                    }
                }
                else
                {
                    //Not checked
                    cksts[loopA] = 0;
                }
                //
                ax = ax << 1;
            }
            //
            int ret = 0;
            for (int loopA = 0; loopA < 4; loopA++)
            {
                if (ret < cksts[loopA])
                {
                    ret = cksts[loopA];
                }
            }
            //
            return (ret);
        }

    } // class
}
