/*
  Motion Puttern Builder
  Sample for Arduino
             ESP32-WROOM / ESP32-WROVER

  Target is PCL6045

  NIPPON PULSE MOTOR Co., Ltd.
*/

#include "MpbSample_data.h"
#include "TargetDevice.h"

extern Mpb_Samples sample;
unsigned long GetMicroSec();

//----------------------------------------------
//Flow controll
//Variables used in WAIT
//Manage time in microseconds
bool WaitEn = false;
unsigned long TsTime = 0;//Holds the initial time used in WAIT
unsigned long WaitT;//wait time
unsigned long NwTime;
unsigned long CpTime;//For comparison
//Used in register operations
unsigned short REG[WorkBuffSize];//Register on Flow
bool ZeroFlag = false;
bool OverFlag = false;


//----------------------------------------------
//Targer device controll
char PreRegStatus; //PreRegister status 0:Not Busy   1:Busy bat Not use PreRegister  2:PreRegister is full
int ExePoint = 0;//Execution point

//----------------------------------------------
//SPI send and receive
#define SpiBuffCnts 256     // Max size of SPI transceiver buffer
extern char SpiBuff[SpiBuffCnts];  // SPI transceiver buffer
extern int SendCnt;                // Number of transfers

void Spi_transfer();


//----------------------------------------------
//Targer device main-status and port
short Msts[4] = {0, 0, 0, 0};
char Port[4] = {0, 0, 0, 0};

//----------------------------------------------
//UNION
union ShortChar
{
  char CH[2];
  short SH;
};
//
union LongChar
{
  char CH[4];
  long LH;
};


///////////////////////////////////////////////////////////////////////////////
// Read status from Targer device
//
void Target_Device::Read_Status()
{
  int DumyCnt = 8;
  SendCnt = 0;
  //
  SpiBuff[SendCnt++] = (sample.Ds << 6) | 0x01 | 0x0F;//All axis
  for (int loopA = 0; loopA < DumyCnt; loopA++)
  {
    SpiBuff[SendCnt++] = 0;//Dumy
  }
  //
  Spi_transfer();
  //
  int GetPoint = 1;
  ShortChar sc;//union
  //
  for (int loopA = 0; loopA < AxisCnt; loopA++)
  {
    sc.CH[0] = SpiBuff[GetPoint++];
    sc.CH[1] = SpiBuff[GetPoint++];
    Msts[loopA] = sc.SH;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Read port from Targer device
//
void Target_Device::Read_Port()
{
  int DumyCnt = 8;
  SendCnt = 0;
  //
  SpiBuff[SendCnt++] = (sample.Ds << 6) | 0x30 | 0x0F;//all axis
  //
  for (int loopA = 0; loopA < DumyCnt; loopA++)
  {
    SpiBuff[SendCnt++] = 0;//Dumy
  }
  //
  Spi_transfer();
  ShortChar sc;//union
  //
  int GetPoint = 1;
  for (int loopA = 0; loopA < AxisCnt; loopA++)
  {
    sc.CH[0] = SpiBuff[GetPoint++]; // port
    sc.CH[1] = SpiBuff[GetPoint++]; // sub-status
    Port[loopA] =  sc.SH;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Check pre-register status
//
// return 0: Device is not busy
//        1: Empty pre-register and Device is Busy
//        2: Pre-register is full
//
int Target_Device::CheckPreReg(char UseAxis)
{
  int ret = 0;
  char ax = 1;
  //
  for(int loopA = 0; loopA < AxisCnt; loopA++)
  {
    if((ax & UseAxis) == ax)
    {
      if((Msts[loopA] & 0x4000) == 0x4000)  // Check pre-register full
      {
        ret = 2;
      }
      else if((Msts[loopA] & 0x4001) == 0x0001)  // Empty pre-register and Device is Busy
      {
        ret = 1;
      }
    }
    //
    ax = ax << 1;
  }
  //
  return(ret);
}


///////////////////////////////////////////////////////////////////////////////
// Checking motion ax1s
//
char Target_Device::CheckAxis(int num)
{
  char ax = 0;
  //
  for(int loopA = 0; loopA < AxisCnt; loopA++)
  {
    if(sample.RegDat[num][loopA][7] != 0)
    {
      ax = ax | (char)(1 << loopA);
    }
  }
  //
  return(ax);
}


///////////////////////////////////////////////////////////////////////////////
// Operating status check
//
int Target_Device::CheckMotionStatus()
{
    Read_Status();
    PreRegStatus = CheckPreReg(0x0F);
    //
    if (PreRegStatus == 0)
    {
        //all actions completed
        return(1);
    }
    else
    {
        return(0);
    }
 }


///////////////////////////////////////////////////////////////////////////////
// 
//
int Target_Device::MotioneProcess(int num)
{
  int ret = 0;

  Read_Status();
  char UseAxis = CheckAxis(num);
  PreRegStatus = CheckPreReg(UseAxis);

  //
  if((PreRegStatus == 0) || //Device is not busy.
     (PreRegStatus == 1) )  //Device is busy but pre-register is empty. 
  {
    LongChar lc;//union

    //
    //Send Register write command.
    for(int loopA = 0; loopA < RegCnt; loopA++)
    {
      char ax = 0;
      //
      SendCnt = 1;
      //Top command set is after all data set
      SpiBuff[SendCnt++] = sample.RegWt[loopA]; // Write command
      //
      //Register write command set to buffer.
      for(int loopB = 0; loopB < AxisCnt; loopB++)
      {
        if(sample.RegDat[num][loopB][7] != 0)  // This axis is use
        {
          ax = ax | (char)(1 << loopB);
          //
          lc.LH = sample.RegDat[num][loopB][loopA];
          SpiBuff[SendCnt++] = lc.CH[0];
          SpiBuff[SendCnt++] = lc.CH[1];
          SpiBuff[SendCnt++] = lc.CH[2];
          SpiBuff[SendCnt++] = lc.CH[3];
        }
      }   
      //
      if(ax != 0)
      {
        //
        SpiBuff[0] = (sample.Ds << 6) | ax; // Top command set
        //
        Spi_transfer();
      }
    }//Send Register write command.
    //
    //Start command set to buffer
    //group same commands 
    char StAx[] = {0, 0, 0, 0};
    char StCom[] = {0x50, 0x51, 0x52, 0x53};
    char ax = 0x01;
    for(int loopA = 0; loopA < AxisCnt; loopA++)
    {
      switch((int)(sample.RegDat[num][loopA][7]))
      {
        case 0x50 ://STAFL
                   StAx[0] = StAx[0] | ax;
                   break;
        case 0x51 ://STAFH
                   StAx[1] = StAx[1] | ax;
                   break;
        case 0x52 ://STUD
                   StAx[2] = StAx[2] | ax;
                   break;
        case 0x53 ://STAUD
                   StAx[3] = StAx[3] | ax;
                   break;
      }
      //
      ax = ax << 1;
    }//group same commands
    //
    //Send start command
    for(int loopA = 0; loopA < 4; loopA++)
    {
      if(StAx[loopA] != 0)
      {
        SendCnt = 0;
        SpiBuff[SendCnt++] = (sample.Ds << 6) | 0x00 | StAx[loopA]; // Top command set
        SpiBuff[SendCnt++] = StCom[loopA]; // Write command
        //
        Spi_transfer();
      }
    }//group same commands   
    //
    ret = 1;
  }
  //
  //ret = 1; //<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Absolutely Delete !!!!!!!!!!!!!!!!
  return(ret);
}


///////////////////////////////////////////////////////////////////////////////
// Write PCL6045-register
//
void Target_Device::WriteReg_hw_ini(char com, short *data)
{
  LongChar lc;//union

  //
  SendCnt = 0;
  SpiBuff[SendCnt++] = (sample.Ds << 6) | 0x0F; // All axis
  SpiBuff[SendCnt++] = sample.RegWt[com]; // Write command
  //
  for(int loopA = 0; loopA < AxisCnt; loopA++)
  {
    lc.LH = *(data + loopA);
    SpiBuff[SendCnt++] = lc.CH[0];
    SpiBuff[SendCnt++] = lc.CH[1];
    SpiBuff[SendCnt++] = lc.CH[2];
    SpiBuff[SendCnt++] = lc.CH[3];
  }
  //
  Spi_transfer();
}


///////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////
// WAIT operation
//
// If any axis are in motion, all axis are stopped and called.
///////////////////////////////////////////////////////////////////////////////////////
int Target_Device::WaitExe(short WtTime)
{
  int ret = 0;//Default is to not proceed

  if (WaitEn == false) //Not in WAIT processing
  {
    Read_Status();
    PreRegStatus = CheckPreReg(0x0F);
    if (PreRegStatus == 0)
    {
      //
      WaitT = WtTime * 1000;//to microseconds
      WaitEn = true;
      //
      TsTime = GetMicroSec();//time at start
    }
  }
  else//WAIT in operation
  {
    NwTime = GetMicroSec();
    if (TsTime > NwTime) //timer overflowed(70 minutes passed)
    {
      NwTime = (0xFFFFFFFF - TsTime)  + NwTime + 1;
    }
    else
    {
      NwTime = NwTime - TsTime;
    }
    //WAIT process end
    if (NwTime >= WaitT) //WAIT end
    {
      //
      WaitEn = false;
      //
      ret = 1;//proceed
    }
  }
  //
  return (ret);
}


///////////////////////////////////////////////////////////////////////////////////////
// Register Subtraction operation
//
//  Comm  : Operation type
//  RegNo : Register number (storage destination)
//  dat   : Data or Data source register number
///////////////////////////////////////////////////////////////////////////////////////
void CheckStattValue(unsigned long RegFlag)
{
  //
  if((RegFlag & 0x0000FFFF) == 0)
  {
    ZeroFlag = true;
  }
  else
  {
    ZeroFlag = false;
  }
  //
  if((RegFlag & 0xFFFF0000) != 0)
  {
    OverFlag = true;
  }
  else
  {
    OverFlag = false;
  }
}


///////////////////////////////////////////////////////////////////////////////////////
// Register Addition operation
//
//  Comm  : Operation type
//  RegNo : Register number (storage destination)
//  dat   : Data or Data source register number
///////////////////////////////////////////////////////////////////////////////////////
unsigned short Target_Device::REG_Addition(unsigned short Reg, int dat)
{
  unsigned long RegFlag;
  //
  RegFlag = (unsigned long)Reg + (unsigned long)dat;
  CheckStattValue(RegFlag);
  //
  return((unsigned short)(RegFlag & 0x0000FFFF));
}


///////////////////////////////////////////////////////////////////////////////////////
// Register Subtraction operation
//
//  Comm  : Operation type
//  RegNo : Register number (storage destination)
//  dat   : Data or Data source register number
///////////////////////////////////////////////////////////////////////////////////////
unsigned short Target_Device::REG_Subtraction(unsigned short Reg, int dat)
{
  unsigned long RegFlag;
  //
  RegFlag = (unsigned long)Reg - (unsigned long)dat;
  CheckStattValue(RegFlag);
  //
  return((unsigned short)(RegFlag & 0x0000FFFF));
}


///////////////////////////////////////////////////////////////////////////////////////
// Stop command transfer operation
//
///////////////////////////////////////////////////////////////////////////////////////
void Target_Device::StopCommand()
{
  //
  SendCnt = 0;
  SpiBuff[SendCnt++] = (sample.Ds << 6) | 0x0F; // All axis
  SpiBuff[SendCnt++] = 0x49; // Write command
  //
  Spi_transfer();
}
