﻿//////////////////////////////////////////////////////////////////////////////
//
//  FTDI access source  ( for PCL6045 )
//
//  NIPPON PULSE MOTOR Co., Ltd.
//
//////////////////////////////////////////////////////////////////////////////

using System;
using System.Runtime.InteropServices;
using FTD2XX_NET;
using System.Threading;
using System.Diagnostics;
using System.Linq;

namespace MP_Builder_Sample
{
    public partial class clsFTDI
    {
        ////////////////////////////////////////////////////////////////////
        /// 
        public FTDI ftHandle = new FTDI();
        //
        public byte Ds = 0;//Device number is changed in software
        //keep status
        public uint[] MSTS = new uint[4];
        public uint[] SSTS = new uint[4];
        public byte[] port = new byte[] { 0, 0, 0, 0 };
        //Maximum definition
        private const int MAX_TAG = 65536;
        private const int MAX_REG = 256;
        private const int MAX_BUFF = 4096;
        //
        public int index = 0;
        public byte[] WriteBuffer = new byte[MAX_BUFF];
        public int indexS = 0;//for status read
        public byte[] WriteBufferS = new byte[MAX_BUFF];//for status read
        public byte[] ReadBuffer = new byte[MAX_BUFF];
        public bool FlowData_on = true;//active flag
        public bool FlowData_stop = false;//Suspend flag = off

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// use as union
        /// </summary>
        [StructLayout(LayoutKind.Explicit)]
        public struct TBytes
        {
            [FieldOffset(0)]
            public byte Byte0;

            [FieldOffset(1)]
            public byte Byte1;

            [FieldOffset(2)]
            public byte Byte2;

            [FieldOffset(3)]
            public byte Byte3;

            [FieldOffset(0)]
            public uint Word;
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// FTDI connection processing
        /// </summary>
        /// <param name="description"></param>
        /// <returns></returns>
        public FTDI.FT_STATUS LSI_Open(Form1 fm1, string description)
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;

            ftStatus = ftHandle.OpenByDescription(description);

            if (!ftHandle.IsOpen)
            {   // error
                Debug.WriteLine("Error: FTDI.penByDescription! [ " + ftStatus.ToString() + " ]");
                return ftStatus;
            }

            ////////////////////////////////////////////////////////////////////
            // FTDI.ResetDevice
            ftStatus = ftHandle.ResetDevice();

            if (ftStatus != FTDI.FT_STATUS.FT_OK)
            {   // error
                Debug.WriteLine("Error: FTDI.ResetDevice! [ " + ftStatus.ToString() + " ]");
                ftHandle.Close();
                return ftStatus;
            }

            Thread.Sleep(100);     //Wair fot reset process

            ////////////////////////////////////////////////////////////////////
            // FTDI.SetLatency (1ms)
            ftStatus = ftHandle.SetLatency(1);

            if (ftStatus != FTDI.FT_STATUS.FT_OK)
            {   // error
                Debug.WriteLine("Error: FTDI.SetLatency! [ " + ftStatus.ToString() + " ]");
                ftHandle.Close();
                return ftStatus;
            }

            ////////////////////////////////////////////////////////////////////
            // FTDI.SetTimeouts
            uint ReadTimeout = 1000;    // 1000ms
            uint WriteTimeout = 1000;   // 1000ms

            ftStatus = ftHandle.SetTimeouts(ReadTimeout, WriteTimeout);

            if (ftStatus != FTDI.FT_STATUS.FT_OK)
            {   // error
                Debug.WriteLine("Error: FTDI.SetTimeouts! [ " + ftStatus.ToString() + " ]");
                ftHandle.Close();
                return ftStatus;
            }

            ////////////////////////////////////////////////////////////////////
            // FTDI.SetBitMode (MCU Host Bus Emulation Mode)
            byte Mask = 0;
            byte BitMode = FTDI.FT_BIT_MODES.FT_BIT_MODE_MCU_HOST;

            ftStatus = ftHandle.SetBitMode(Mask, BitMode);

            if (ftStatus != FTDI.FT_STATUS.FT_OK)
            {   // error
                Debug.WriteLine("Error: FTDI.SetBitMode! [ " + ftStatus.ToString() + " ]");
                ftHandle.Close();
                return ftStatus;
            }

            ////////////////////////////////////////////////////////////////////
            // FTDI.Write (12MHz -> 60MHz)
            byte[] dataBuffer = new byte[] { 0x8A };    // 60MHz
            int numBytesToWrite = 1;
            uint numBytesWritten = 0;

            ftStatus = ftHandle.Write(dataBuffer, numBytesToWrite, ref numBytesWritten);

            if (ftStatus != FTDI.FT_STATUS.FT_OK)
            {   // error
                Debug.WriteLine("Error: FTDI.Write! [ " + ftStatus.ToString() + " ]");
                return ftStatus;
            }

            ////////////////////////////////////////////////////////////////////
            // FTDI.Purge
            //  Clear transmit and receive buffers
            ftStatus = ftHandle.Purge(FTDI.FT_PURGE.FT_PURGE_RX | FTDI.FT_PURGE.FT_PURGE_TX);

            if (ftStatus != FTDI.FT_STATUS.FT_OK)
            {   // error
                Debug.WriteLine("Error: FTDI.Purge! [ " + ftStatus.ToString() + " ]");
                ftHandle.Close();
                return ftStatus;
            }

            ////////////////////////////////////////////////////////////////////
            return ftStatus;
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public FTDI.FT_STATUS LSI_Close()
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;

            ////////////////////////////////////////////////////////////////////
            // FTDI.Close
            ftStatus = ftHandle.Close();

            if (ftHandle.IsOpen)
            {   // error
                Debug.WriteLine("Error: FTDI.Close! [" + ftStatus.ToString() + "]");
                return ftStatus;
            }

            return ftStatus;
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// FTDI Write
        /// </summary>
        /// <param name="dataBuffer"></param>
        /// <param name="numBytesToWrite"></param>
        /// <returns></returns>
        public FTDI.FT_STATUS LSI_Write(byte[] dataBuffer, int numBytesToWrite)
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;

            ////////////////////////////////////////////////////////////////////
            // FTDI.Write
            uint numBytesWritten = 0;

            ftStatus = ftHandle.Write(dataBuffer, numBytesToWrite, ref numBytesWritten);

            ////////////////////////////////////////////////////////////////////
            // Error Check
            if (ftStatus != FTDI.FT_STATUS.FT_OK)
            {   // error
                Debug.WriteLine("Error: FTDI.Write! [ " + ftStatus.ToString() + " ]");
                return ftStatus;
            }

            if (numBytesToWrite != numBytesWritten)
            {   // error
                Debug.WriteLine(
                    "Error: FTDI.Write! [ numBytesToWrite: " + numBytesToWrite.ToString() +
                    " /= numBytesWritten: " + numBytesWritten.ToString() + " ]");
                return FTDI.FT_STATUS.FT_OTHER_ERROR;
            }

            ////////////////////////////////////////////////////////////////////
            return ftStatus;
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Garbage may remain after forced termination.
        /// This function eliminates garbage.
        /// </summary>
        /// <param name="dataBuffer"></param>
        /// <returns></returns>
        public uint LSI_CheckRead(ref byte[] dataBuffer)
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;
            uint RxQueue = 0;
            uint Dumy = 0;

            ftStatus = ftHandle.GetRxBytesAvailable(ref RxQueue);
            //
            if ((ftStatus == FTDI.FT_STATUS.FT_OK) && (RxQueue > 0))
            {
                ftStatus = ftHandle.Read(dataBuffer, RxQueue, ref Dumy);
                RxQueue = Dumy;
            }
            //
            return(RxQueue);
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// FTDI Read(Hight speed)
        /// </summary>
        /// <param name="dataBuffer"></param>
        /// <param name="numBytesRead"></param>
        /// <returns></returns>
        public FTDI.FT_STATUS LSI_Read_first(ref byte[] dataBuffer, ref uint numBytesRead)
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;

            ////////////////////////////////////////////////////////////////////
            // FTDI.Read

            int bad = 0;
            uint RxQueue = 0;
            do
            {
                ftStatus = ftHandle.GetRxBytesAvailable(ref RxQueue);
                bad++;
                if (bad > 5000)
                {
                    ftStatus = FTDI.FT_STATUS.FT_EEPROM_READ_FAILED;
                    break;
                }

            } while (RxQueue < numBytesRead);

            uint numBytesToRead = 0;
            do
            {
                ftStatus = ftHandle.Read(dataBuffer, numBytesRead, ref numBytesToRead);

                if (ftStatus != FTDI.FT_STATUS.FT_OK)
                {   // error
                    Debug.WriteLine("Error: FTDI.Read! [" + ftStatus.ToString() + "]");
                    return ftStatus;
                }
                //
                numBytesRead = numBytesRead - numBytesToRead;
            }
            while (numBytesRead > 0);

            return ftStatus;
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Get axis address from axis select command
        /// If multiple axes are specified, the lower side takes precedence
        ///   Jsc  ： Axis selection command
        ///   ofs  ： address offset
        /// </summary>
        /// <param name="description"></param>
        /// <returns></returns>
        public byte GetAdd(byte Jsc, byte ofs)
        {
            byte ret = 0;
            //
            if ((Jsc & 0x02) != 0)
            {
                ret = 0x08;
            }
            else if ((Jsc & 0x04) != 0)
            {
                ret = 0x10;
            }
            else if ((Jsc & 0x08) != 0)
            {
                ret = 0x18;
            }
            //
            return ((byte)(ret + ofs));
        }


        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Set buffer the command Note that only one axis is processed
        ///   Jsc  ： Axis selection command
        ///   comm ： register write command
        ///   RegD ： write data
        /// </summary>
        /// <returns></returns>
        public void Write_REG(byte Jsc, byte comm, uint RegD)
        {
            var bytes = new TBytes();
            byte add = GetAdd(Jsc, 4);
            //
            bytes.Word = RegD;
            //
            WriteBuffer[index++] = 0x93; // FTDI-write command & address & data
            WriteBuffer[index++] = (byte)(add++); // AddrHigh(0h) bit 7:0
            WriteBuffer[index++] = 0x00; // AddrLow(Dummy)
            WriteBuffer[index++] = bytes.Byte0;//Data 7-0
            //
            WriteBuffer[index++] = 0x93; // FTDI-write command & address & data
            WriteBuffer[index++] = (byte)(add++); // AddrHigh(0h) bit 15:8
            WriteBuffer[index++] = 0x00; // AddrLow(Dummy)
            WriteBuffer[index++] = bytes.Byte1;//Data 15-8
            //
            WriteBuffer[index++] = 0x93; // FTDI-write command & address & data
            WriteBuffer[index++] = (byte)(add++); // AddrHigh(0h) bit 24:16
            WriteBuffer[index++] = 0x00; // AddrLow(Dummy)
            WriteBuffer[index++] = bytes.Byte2;//Data 23-16
            //
            WriteBuffer[index++] = 0x93; // FTDI-write command & address & data
            WriteBuffer[index++] = add; // AddrHigh(0h) bit 31:24
            WriteBuffer[index++] = 0x00; // AddrLow(Dummy)
            WriteBuffer[index++] = bytes.Byte3;//Data 31-24
            //
            // command
            add = (byte)(add - 6);
            //
            WriteBuffer[index++] = 0x93; // FTDI-write command & address
            WriteBuffer[index++] = (byte)(add--); // AddrHigh
            WriteBuffer[index++] = 0x00; // AddrLow(Dummy)
            WriteBuffer[index++] = Jsc; //select axis
            //
            WriteBuffer[index++] = 0x93; // FTDI-write command & address
            WriteBuffer[index++] = add; // AddrHigh
            WriteBuffer[index++] = 0x00; // AddrLow(Dummy)
            WriteBuffer[index++] = comm;
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Set buffer the command Note that only one axis is processed
        ///   Jsc  ： Axis select command (Not use)
        ///   comm ： command
        /// </summary>
        /// <returns></returns>
        public void Write_COM(byte Jsc, byte comm)
        {
            byte add = GetAdd(Jsc, 1);
            //
            WriteBuffer[index++] = 0x93; // FTDI-write command & address
            WriteBuffer[index++] = (byte)(add--); // AddrHigh
            WriteBuffer[index++] = 0x00; // AddrLow(Dummy)
            WriteBuffer[index++] = Jsc; //Axis selection is fixed to X
            //
            WriteBuffer[index++] = 0x93; // FTDI-write command & address
            WriteBuffer[index++] = add; // AddrHigh
            WriteBuffer[index++] = 0x00; // AddrLow(Dummy)
            WriteBuffer[index++] = comm;
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Send seted data via USB
        /// </summary>
        /// <returns></returns>
        public FTDI.FT_STATUS SendUsb(byte[] WriteBuffer, ref int index)
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;
            //
            if (index > 0)
            {
                ftStatus = LSI_Write(WriteBuffer, index);
                index = 0;
            }
            //
            return (ftStatus);
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Receive data from USB transmission
        ///  ReadBuffer : number of bytes to read
        ///  numBytesRead : number of bytes read
        /// </summary>
        /// <returns></returns>
        private FTDI.FT_STATUS ReadUsb(ref byte[] ReadBuffer, ref uint numBytesRead)
        {
            ////////////////////////////////////////////////////////////////////
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;
            //
            ftStatus = LSI_Read_first(ref ReadBuffer, ref numBytesRead);
            //
            return (ftStatus);
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Set the status read commans
        /// </summary>
        public void Read_Status(int AxisConts)
        {
            //
            indexS = 0;
            //
            for (int loops = 0; loops < AxisConts; loops++)
            {
                byte add = (byte)(loops << 3);
                //
                WriteBufferS[indexS++] = 0x91; // FTDI-read command & address
                WriteBufferS[indexS++] = (byte)(add++); // AddrHigh(0h) MainStatus-low
                WriteBufferS[indexS++] = 0x00; // AddrLow(Dummy)
                //
                WriteBufferS[indexS++] = 0x91; // FTDI-read command & address
                WriteBufferS[indexS++] = (byte)(add++); // AddrHigh(1h) MainStatus-High
                WriteBufferS[indexS++] = 0x00; // AddrLow(Dummy)
                //
                WriteBufferS[indexS++] = 0x91; // FTDI-read command & address
                WriteBufferS[indexS++] = (byte)(add++); // AddrHigh(2h) PORT
                WriteBufferS[indexS++] = 0x00; // AddrLow(Dummy)
                //
                WriteBufferS[indexS++] = 0x91; // FTDI-read command & address
                WriteBufferS[indexS++] = add; // AddrHigh(3h) SubStatus
                WriteBufferS[indexS++] = 0x00; // AddrLow(Dummy)
            }
            //
            WriteBufferS[indexS++] = 0x87;
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Pick up status and sub-status from USB read buffer "ReadBuffer"
        ///   ofset ： offset to the beginning of the data (not used in this source, just match the other arguments)
        ///   MSTS ： Main status read result
        ///   SSTS  ： Substatus read result
        /// </summary>
        /// <returns></returns>
        public void PickUpStatus(Form1 fm1, int ofset)
        {
            var bytes = new TBytes();
            //
            bytes.Byte2 = (byte)0;//Data 23-16
            bytes.Byte3 = (byte)0;//Data 31-24
            //
            for (int loops = 0, of1 = 0; loops < fm1.hDataSmp.AxisConts; loops++)
            {
                of1 = loops * 4 + ofset;
                //main status
                bytes.Byte0 = (byte)(ReadBuffer[of1]);    //Data 7-0
                bytes.Byte1 = (byte)(ReadBuffer[of1 + 1]);//Data 15-8
                MSTS[loops] = (uint)(bytes.Word);
                //syb status & port
                bytes.Byte0 = (byte)(ReadBuffer[of1 + 2]);//Data 7-0
                bytes.Byte1 = (byte)(ReadBuffer[of1 + 3]);//Data 15-8
                SSTS[loops] = (uint)(bytes.Word);
                //port
                port[loops] = (byte)(ReadBuffer[of1 + 2]);
            }
        }

        ////////////////////////////////////////////////////////////////////
        /// <summary>
        /// Wait for operation completion
        /// Do not move to the next operation until all of the specified axes are completed or the pre-registration is free
        ///
        ///  AxisConts : Number of LSI axes
        ///  UseAxis   : Specify axis to wait for status
        ///
        ///  return    : 0＝All specified axes are stopped
        ///            : 1＝All specified axes are in motion but pre-registration is empty or stopped
        ///            : 3＝One of the specified axes is in motion and pre-registration is not free
        /// </summary>
        public int WaitMsts(Form1 fm1, int AxisConts, int UseAxis)
        {
            FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;
            int[] cksts = new int[] {3, 3, 3, 3};
            //
            Read_Status(AxisConts);
            //
            ftStatus = SendUsb(WriteBufferS, ref indexS);
            //
            if (ftStatus == FTDI.FT_STATUS.FT_OK)
            {
                uint numBytesRead = (uint)(AxisConts * 4);
                ftStatus = ReadUsb(ref ReadBuffer, ref numBytesRead);
                //
                PickUpStatus(fm1, 0);
            }
            //
            for(int loopA = 0, ax = 1; loopA < 4; loopA++)
            {
                if((UseAxis & ax) == ax)
                {
                    if(((MSTS[loopA] & 0x0000400F) == 0x00000000) ||  // Immediately after reset
                       ((MSTS[loopA] & 0x0000400F) == 0x00000008))    // stop working
                    {
                        //stopped
                        cksts[loopA] = 0;
                    }
                    else
                    {
                       if((MSTS[loopA] & 0x00004000) == 0)
                        {
                            //In operation, pre-register is empty
                            cksts[loopA] = 1;
                        }
                    }
                }
                else
                {
                    //Not checked
                    cksts[loopA] = 0;
                }
                //
                ax = ax << 1;
            }
            //
            int ret = 0;
            for(int loopA = 0; loopA < 4; loopA++)
            {
                if(ret < cksts[loopA])
                {
                    ret = cksts[loopA];
                }
            }
            return(ret);
        }

    } // class
}
